#include <stdio.h>
#include <stdlib.h>

#include "../include/GL/glew.h"

#include "../include/buffers.h"
#include "../include/global.h"

namespace	VMI
{
Vertex_ *buf_vertices = NULL;
Color   *buf_colors = NULL;

Vertex_ *pbuf_vertices;
Color   *pbuf_colors;

GLuint vertex_buf = 0,
       color_buf = 0;

}

using namespace	VMI;

void VMI::saveVertexBuffer(Change *c, VertexIL *dest) {
    int j, t, n = 0, m;
    
    // Save vertex coordinates and color for deleted triangles
    for (j=0; j<c->numDel; j++) { 
        
        t = c->deleted[j].id;
        
        m = 3 * t;
        
        dest[n].x = buf_vertices[m + 0].x;
        dest[n].y = buf_vertices[m + 0].y;
        dest[n].z = buf_vertices[m + 0].z;
        dest[n].a = buf_colors[m + 0].a;
        dest[n].b = buf_colors[m + 0].b;
        dest[n].r = buf_colors[m + 0].r;
        dest[n].g = buf_colors[m + 0].g;
        n++;
        
        dest[n].x = buf_vertices[m + 1].x;
        dest[n].y = buf_vertices[m + 1].y;
        dest[n].z = buf_vertices[m + 1].z;
        dest[n].a = buf_colors[m + 1].a;
        dest[n].b = buf_colors[m + 1].b;
        dest[n].r = buf_colors[m + 1].r;
        dest[n].g = buf_colors[m + 1].g;
        n++;
        
        dest[n].x = buf_vertices[m + 2].x;
        dest[n].y = buf_vertices[m + 2].y;
        dest[n].z = buf_vertices[m + 2].z;
        dest[n].a = buf_colors[m + 2].a;
        dest[n].b = buf_colors[m + 2].b;
        dest[n].r = buf_colors[m + 2].r;
        dest[n].g = buf_colors[m + 2].g;
        n++;
    }
    // Save vertex coordinates for modified triangles
    for (j=0; j<c->numMod; j++) { 
        
        t = c->modified[j].id;
        
        m = 3 * t;
        
        dest[n].x = buf_vertices[m + 0].x;
        dest[n].y = buf_vertices[m + 0].y;
        dest[n].z = buf_vertices[m + 0].z;
        n++;
        
        dest[n].x = buf_vertices[m + 1].x;
        dest[n].y = buf_vertices[m + 1].y;
        dest[n].z = buf_vertices[m + 1].z;
        n++;
        
        dest[n].x = buf_vertices[m + 2].x;
        dest[n].y = buf_vertices[m + 2].y;
        dest[n].z = buf_vertices[m + 2].z;
        n++;
    }
}

void VMI::loadVertexBuffer(VertexIL *src, Change *c) {
    int j, t, n = 0, m;
    
    // Load vertex coordinates and color for deleted triangles
    for (j=0; j<c->numDel; j++) { 
        
        t = c->deleted[j].id;
        
        m = 3 * t;

        buf_vertices[m + 0].x = src[n].x;
        buf_vertices[m + 0].y = src[n].y;
        buf_vertices[m + 0].z = src[n].z;
        buf_colors[m + 0].a = src[n].a;
        buf_colors[m + 0].b = src[n].b;
        buf_colors[m + 0].r = src[n].r;
        buf_colors[m + 0].g = src[n].g;
        n++;

        buf_vertices[m + 1].x = src[n].x;
        buf_vertices[m + 1].y = src[n].y;
        buf_vertices[m + 1].z = src[n].z;
        buf_colors[m + 1].a = src[n].a;
        buf_colors[m + 1].b = src[n].b;
        buf_colors[m + 1].r = src[n].r;
        buf_colors[m + 1].g = src[n].g;
        n++;

        buf_vertices[m + 2].x = src[n].x;
        buf_vertices[m + 2].y = src[n].y;
        buf_vertices[m + 2].z = src[n].z;
        buf_colors[m + 2].a = src[n].a;
        buf_colors[m + 2].b = src[n].b;
        buf_colors[m + 2].r = src[n].r;
        buf_colors[m + 2].g = src[n].g;
        n++;
    }
    // Load vertex coordinates for modified triangles
    for (j=0; j<c->numMod; j++) { 
        
        t = c->modified[j].id;
        
        m = 3 * t;

        buf_vertices[m + 0].x = src[n].x;
        buf_vertices[m + 0].y = src[n].y;
        buf_vertices[m + 0].z = src[n].z;
        n++;

        buf_vertices[m + 1].x = src[n].x;
        buf_vertices[m + 1].y = src[n].y;
        buf_vertices[m + 1].z = src[n].z;
        n++;

        buf_vertices[m + 2].x = src[n].x;
        buf_vertices[m + 2].y = src[n].y;
        buf_vertices[m + 2].z = src[n].z;
        n++;
    }
}

void VMI::setupVertexArray(Mesh *mesh, Color *colors)
{
    int i, v1, v2, v3, n = 0;
    GLubyte r, g, b, a;

    if (buf_vertices != NULL) free(buf_vertices);
    if (buf_colors != NULL) free(buf_colors);

    buf_vertices = (Vertex_ *)malloc(mesh->numTriangles * 3 * sizeof(Vertex_));

    buf_colors = (Color *)malloc(mesh->numTriangles * 3 * sizeof(Color));

    if (buf_vertices == NULL) {
        fprintf(stderr, "Error allocating memory\n");
        exit(1);
    }

    if (buf_colors == NULL) {
        fprintf(stderr, "Error allocating memory\n");
        exit(1);
    }

    for (i=0; i<mesh->numTriangles; i++) {

        v1= mesh->triangles[i].indices[0];
        v2= mesh->triangles[i].indices[1];
        v3= mesh->triangles[i].indices[2];
        
        if (mesh->triangles[i].enable == TRUE) {
            
            r = colors[i].r;
            g = colors[i].g;
            b = colors[i].b;
            a = colors[i].a;
            
            buf_colors[n].r = r;
            buf_colors[n].g = g;
            buf_colors[n].b = b;
            buf_colors[n].a = a;
            buf_vertices[n].x = mesh->vertices[v1].x;
            buf_vertices[n].y = mesh->vertices[v1].y;
            buf_vertices[n].z = mesh->vertices[v1].z;
            n++;
            
            buf_colors[n].r = r;
            buf_colors[n].g = g;
            buf_colors[n].b = b;
            buf_colors[n].a = a;
            buf_vertices[n].x = mesh->vertices[v2].x;
            buf_vertices[n].y = mesh->vertices[v2].y;
            buf_vertices[n].z = mesh->vertices[v2].z;
            n++;
            
            buf_colors[n].r = r;
            buf_colors[n].g = g;
            buf_colors[n].b = b;
            buf_colors[n].a = a;
            buf_vertices[n].x = mesh->vertices[v3].x;
            buf_vertices[n].y = mesh->vertices[v3].y;
            buf_vertices[n].z = mesh->vertices[v3].z;
            n++;
        } else {
         
            buf_colors[n].r = 0;
            buf_colors[n].g = 0;
            buf_colors[n].b = 0;
            buf_colors[n].a = 0;
            buf_vertices[n].x = 0;
            buf_vertices[n].y = 0;
            buf_vertices[n].z = 0;
            n++;
            
            buf_colors[n].r = 0;
            buf_colors[n].g = 0;
            buf_colors[n].b = 0;
            buf_colors[n].a = 0;
            buf_vertices[n].x = 0;
            buf_vertices[n].y = 0;
            buf_vertices[n].z = 0;
            n++;
            
            buf_colors[n].r = 0;
            buf_colors[n].g = 0;
            buf_colors[n].b = 0;
            buf_colors[n].a = 0;
            buf_vertices[n].x = 0;
            buf_vertices[n].y = 0;
            buf_vertices[n].z = 0;
            n++;
        }
    }
#ifdef VERTEX_ARRAY
    // Enable the buf_vertices vector
    glEnableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);
    
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, buf_colors);
    glVertexPointer(3, GL_FLOAT, 0, buf_vertices);
#endif

    pbuf_colors = buf_colors;
    pbuf_vertices = buf_vertices;
}

void VMI::setupVertexBufferObjects(Mesh *mesh, Color *colors) {
    
    setupVertexArray(mesh, colors);
    
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buf);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, mesh->numTriangles * 3 * 3 * sizeof(float), buf_vertices, GL_STATIC_DRAW_ARB); //upload data

    glVertexPointer(3, GL_FLOAT, 0,0);
    
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, color_buf);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, mesh->numTriangles * 3 * sizeof(Color), buf_colors, GL_STATIC_DRAW_ARB); //upload data
    
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, 0);

    // Enable the vertex array functionality:
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    
    //free(buf_vertices);
    //buf_vertices = NULL;
    
    //free(buf_colors);
    //buf_colors = NULL;
}

void VMI::updateVertexArray(Mesh *mesh, Change *c)
{
    int i, v1, v2, v3, t, n;

    // The deleted triangles are now in background color
    for (i=0; i<c->numDel; i++) {
        t = c->deleted[i].id;

        n = t * 3;

        buf_colors[n + 0].r = 0;
        buf_colors[n + 0].g = 0;
        buf_colors[n + 0].b = 0;
        buf_colors[n + 0].a = 0;

        buf_colors[n + 1].r = 0;
        buf_colors[n + 1].g = 0;
        buf_colors[n + 1].b = 0;
        buf_colors[n + 1].a = 0;

        buf_colors[n + 2].r = 0;
        buf_colors[n + 2].g = 0;
        buf_colors[n + 2].b = 0;
        buf_colors[n + 2].a = 0;

        buf_vertices[n + 0].x = 0;
        buf_vertices[n + 0].y = 0;
        buf_vertices[n + 0].z = 0;

        buf_vertices[n + 1].x = 0;
        buf_vertices[n + 1].y = 0;
        buf_vertices[n + 1].z = 0;

        buf_vertices[n + 2].x = 0;
        buf_vertices[n + 2].y = 0;
        buf_vertices[n + 2].z = 0;
    }

    // Update vertex coordinates for modified triangles
    for (i=0; i<c->numMod; i++) {
        t = c->modified[i].id;

        n = t * 3;

        v1= mesh->triangles[t].indices[0];
        v2= mesh->triangles[t].indices[1];
        v3= mesh->triangles[t].indices[2];

        buf_vertices[n + 0].x = mesh->vertices[v1].x;
        buf_vertices[n + 0].y = mesh->vertices[v1].y;
        buf_vertices[n + 0].z = mesh->vertices[v1].z;

        buf_vertices[n + 1].x = mesh->vertices[v2].x;
        buf_vertices[n + 1].y = mesh->vertices[v2].y;
        buf_vertices[n + 1].z = mesh->vertices[v2].z;

        buf_vertices[n + 2].x = mesh->vertices[v3].x;
        buf_vertices[n + 2].y = mesh->vertices[v3].y;
        buf_vertices[n + 2].z = mesh->vertices[v3].z;
    }
}

void VMI::updateVertexBufferObjects(Mesh *mesh, Change *c) {
    
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buf);
    // Map the buffer object
	buf_vertices = ( Vertex_ *)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY);

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, color_buf);
    // Map the buffer object
	buf_colors = (Color   *)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY);

    updateVertexArray(mesh, c);

    // Unmap buffers
    glBindBuffer(GL_ARRAY_BUFFER_ARB, vertex_buf);
    if (!glUnmapBufferARB(GL_ARRAY_BUFFER_ARB)) {
        // Handle error case
    }
    glBindBuffer(GL_ARRAY_BUFFER_ARB, color_buf);
    if (!glUnmapBufferARB(GL_ARRAY_BUFFER_ARB)) {
        // Handle error case
    }
}

VertexIL *VMI::setupInterleave(Mesh *mesh, Color *colors)
{
    int i, v1, v2, v3, n = 0;
    GLubyte r, g, b, a;
    VertexIL *interleave = NULL;
    
    interleave = (VertexIL *)malloc(mesh->currentNumTriangles * 3 * sizeof(VertexIL));
    
    if (interleave == NULL) {
        fprintf(stderr, "Error allocating memory\n");
        exit(1);
    }
    
    for (i=0; i<mesh->numTriangles; i++) {
        
        if (mesh->triangles[i].enable == TRUE) {
            v1= mesh->triangles[i].indices[0];
            v2= mesh->triangles[i].indices[1];
            v3= mesh->triangles[i].indices[2];
            
            r = colors[i].r;
            g = colors[i].g;
            b = colors[i].b;
            a = colors[i].a;
            
            interleave[n].tx = 0.0f;
            interleave[n].ty = 0.0f;
            interleave[n].r = r;
            interleave[n].g = g;
            interleave[n].b = b;
            interleave[n].a = a;
            interleave[n].x = mesh->vertices[v1].x;
            interleave[n].y = mesh->vertices[v1].y;
            interleave[n].z = mesh->vertices[v1].z;
            n++;
            
            interleave[n].tx = 0.0f;
            interleave[n].ty = 0.0f;
            interleave[n].r = r;
            interleave[n].g = g;
            interleave[n].b = b;
            interleave[n].a = a;
            interleave[n].x = mesh->vertices[v2].x;
            interleave[n].y = mesh->vertices[v2].y;
            interleave[n].z = mesh->vertices[v2].z;
            n++;
            
            interleave[n].tx = 0.0f;
            interleave[n].ty = 0.0f;
            interleave[n].r = r;
            interleave[n].g = g;
            interleave[n].b = b;
            interleave[n].a = a;
            interleave[n].x = mesh->vertices[v3].x;
            interleave[n].y = mesh->vertices[v3].y;
            interleave[n].z = mesh->vertices[v3].z;
            n++;
        }
    }

    return interleave;
}

void VMI::printInterleave(VertexIL *interleave, int num)
{
    int i;

    printf("\n");
    for (i=0; i<num * 3; i++) {
        printf("Interleave%d (%d, %d, %d ,%d, %f, %f ,%f)\n", i,
            interleave[i].r, interleave[i].g, interleave[i].b, interleave[i].a,
            interleave[i].x , interleave[i].y, interleave[i].z);
    }
}

void VMI::updateAllColorInterleave(VertexIL *interleave, int num, Color *colors)
{
    int i, n = 0;
    GLubyte r, g, b, a;

    for (i=1; i<=num; i++) { // Triangles start at 1

        r = colors[i].r;
        g = colors[i].g;
        b = colors[i].b;
        a = colors[i].a;

        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;
        n++;

        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;
        n++;

        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;
        n++;
    }
}

void VMI::updateColorInterleave(VertexIL *interleave, int num, Color *colors, int begin, int end)
{
    int i, n;
    GLubyte r, g, b, a;

    if (end > num) return;

    for (i=begin; i<end; i++) {
        
        n = i * 3;

        r = colors[i].r;
        g = colors[i].g;
        b = colors[i].b;
        a = colors[i].a;

        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;

        n++;
        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;

        n++;
        interleave[n].r = r;
        interleave[n].g = g;
        interleave[n].b = b;
        interleave[n].a = a;
    }
}

void VMI::fillColorInterleave(VertexIL *interleave, int num, int begin, int end, GLubyte color)
{
    int i, n;

    if (end > num) return;

    for (i=begin; i<end; i++) {
        
        n = i * 3;
        interleave[n].r = color;
        interleave[n].g = color;
        interleave[n].b = color;
        interleave[n].a = color;

        n++;
        interleave[n].r = color;
        interleave[n].g = color;
        interleave[n].b = color;
        interleave[n].a = color;

        n++;
        interleave[n].r = color;
        interleave[n].g = color;
        interleave[n].b = color;
        interleave[n].a = color;
    }
}
